// -*- mode: C++; indent-tabs-mode: nil; c-basic-offset: 2; -*-
// Font.cc for Blackbox - an X11 Window manager
// Copyright (c) 2001 - 2002 Sean 'Shaleh' Perry <shaleh@debian.org>
// Copyright (c) 1997 - 2000 Brad Hughes (bhughes@tcac.net)
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#ifdef    HAVE_CONFIG_H
#  include "../config.h"
#endif // HAVE_CONFIG_H

extern "C" {
#ifdef HAVE_STDLIB_H
#  include <stdlib.h>
#endif // HAVE_STDLIB_H
}

#include <iostream>
#include <algorithm>
#include <string>

using std::string;
using std::cerr;
using std::endl;

#include "i18n.hh"
#include "Font.hh"
#include "Util.hh"
#include "GCCache.hh"
#include "Color.hh"

string      BFont::_fallback_font   = "fixed";

#ifdef XFT
BFont::BFont(Display *d, BScreen *screen, const string &family, int size,
             bool bold, bool italic, bool shadow, unsigned char offset, 
             int tint, bool antialias) :
                                          _display(d),
                                          _screen(screen),
                                          _family(family),
                                          _simplename(False),
                                          _size(size),
                                          _bold(bold),
                                          _italic(italic),
                                          _antialias(antialias),
                                          _shadow(shadow),
                                          _offset(offset),
                                          _tint(tint),
                                          _xftfont(0),
                                          _fontset(0),
                                          _fontset_extents(0) {
  XftPattern *pattern;
  
  _valid = False;

  pattern = XftNameParse(_family.c_str());
  
  double patternSize;
  if (XftPatternGetDouble(pattern, "size", 0, &patternSize) != FcResultMatch) {
    XftPatternAddDouble(pattern, "size", _size);
  }

  int weight;
  if (XftPatternGetInteger(pattern, "weight", 0, &weight) != FcResultMatch) {
    if (_bold)
      XftPatternAddInteger(pattern, "weight", 200);
  }

  int slant;
  if (XftPatternGetInteger(pattern, "slant", 0, &slant) != FcResultMatch) {
    if (_italic)
      XftPatternAddInteger(pattern,  "slant", 10);
  }

  pattern = XftFontMatch(_display, _screen->getScreenNumber(), pattern, NULL);
  _xftfont = XftFontOpenPattern(_display, pattern);
  
  if (! _xftfont)
    return;

  _valid = True;
}
#endif


BFont::BFont(Display *d, BScreen *screen, const string &xlfd) :
                                       _display(d),
                                       _screen(screen),
#ifdef    XFT
                                       _antialias(False),
                                       _shadow(False),
                                       _xftfont(0),
#endif // XFT
                                       _fontset(0),
                                       _fontset_extents(0) {
  string int_xlfd;
  if (xlfd.empty())
    int_xlfd = _fallback_font;
  else
    int_xlfd = xlfd;
  
  if ((_valid = createXFont(int_xlfd)))
    return; // success

  if (int_xlfd != _fallback_font) {
    // try the fallback
    cerr << "BFont::BFont(): couldn't load font '" << xlfd << "'" << endl <<
      "Falling back to default '" << _fallback_font << "'" << endl;

    if ((_valid = createXFont(_fallback_font)))
      return; // success
  }

  cerr << "BFont::BFont(): couldn't load font '" << _family << "'" << endl <<
    "Giving up!" << endl;
  return; // failure
}


bool BFont::createXFont(const std::string &xlfd) {
  /*
     This is where _simplename, _bold, _italic, and _size are initialized, since
     they are not initialized in the constructor. This needs to occur before
     calling any Xlfd-building functions.
  */
  if (! parseXlfd(xlfd))
    return False;

  char **missing, *def = "-";
  int nmissing;

  _fontset = XCreateFontSet(_display, buildMultibyteXlfd(xlfd).c_str(),
                            &missing, &nmissing, &def);
  if (nmissing) XFreeStringList(missing);
  if (_fontset) {
    _fontset_extents = XExtentsOfFontSet(_fontset);
    assert(_fontset_extents);
    return True;    
  }
  return False;
}


BFont::~BFont(void) {
#ifdef    XFT
  if (_xftfont)
    XftFontClose(_display, _xftfont);
#endif // XFT

  if (_fontset)
    XFreeFontSet(_display, _fontset);
}

/*
 * Takes _family, _size, _bold, _italic, etc and builds them into a full XLFD.
 */
string BFont::buildMultibyteXlfd(const std::string &xlfd) const {
  string weight = _bold ? "bold" : "medium";
  string slant = _italic ? "i" : "r";
  string sizestr= _size ? itostring(_size) : "*";

  return xlfd + ','
    + "-*-" + _family + "-" +  weight + "-" + slant + "-*-*-" + sizestr + "-*-*-*-*-*-*-*" + ','
    + "-*-*-" + weight + "-" + slant + "-*-*-" + sizestr + "-*-*-*-*-*-*-*" + ','
    + "-*-*-*-*-*-*-" + sizestr + "-*-*-*-*-*-*-*" + ','
    + "*";
}


/*
 * Takes a full X font name and parses it out so we know if we're bold, our
 * size, etc.
 */
bool BFont::parseXlfd(const string &xlfd) {
  if (xlfd.empty() || xlfd[0] != '-') {
    _family = xlfd;
    _simplename = True;
    _bold = False;
    _italic = False;
    _size = 0;
  } else {
    _simplename = False;
    string weight,
           slant,
           sizestr;
    int i = 0;

    string::const_iterator it = xlfd.begin(), end = xlfd.end();
    while(1) {
      string::const_iterator tmp = it;   // current string.begin()
      it = std::find(tmp, end, '-');     // look for comma between tmp and end
      if (i == 2) _family = string(tmp, it); // s[tmp:it]
      if (i == 3) weight = string(tmp, it);
      if (i == 4) slant = string(tmp, it);
      if (i == 7 && string(tmp, it) != "*") sizestr = string(tmp, it);
      if (sizestr.empty() &&
          i == 8 && string(tmp, it) != "*") sizestr = string(tmp, it);
      if (it == end || i >= 8)
        break;
      ++it;
      ++i;
    }
    if (i < 3)  // no name even! can't parse that
      return False;
    _bold = weight == "bold" || weight == "demibold";
    _italic = slant == "i" || slant == "o";
    _size = atoi(sizestr.c_str());
  }
  
  // min/max size restrictions for sanity, but 0 is the font's "default size"
  if (_size && _size < 3)
    _size = 3;
  else if (_size > 97)
    _size = 97;

  return True;
}


void BFont::drawString(Drawable d, int x, int y, const BColor &color,
                       const string &string, bool utf) const {
  assert(_valid);

#ifdef    XFT
  if (_xftfont) {
    XftDraw *draw = XftDrawCreate(_display, d, _screen->getVisual(),
                                  _screen->getColormap());
    assert(draw);

    if (_shadow) {
      XftColor c;
      if (_tint >= 0) {
        c.color.red = 0;
        c.color.green = 0;
        c.color.blue = 0;
        c.color.alpha = 0xffff * _tint/100; // transparent shadow
        c.pixel = BlackPixel(_display, _screen->getScreenNumber());
      } else {
        c.color.red = 0xffff * -_tint/100;
        c.color.green = 0xffff * -_tint/100;
        c.color.blue = 0xffff * -_tint/100;
        c.color.alpha = 0xffff * -_tint/100;
        c.pixel = WhitePixel(_display, _screen->getScreenNumber());
      }

      if (utf) {
        // the string we want to draw is in UTF-8.
#ifdef XFT_UTF8
        // draw it as is with XftDrawStringUtf8()
        XftDrawStringUtf8(draw, &c, _xftfont, x + _offset,
                          _xftfont->ascent + y + _offset,
                          (XftChar8 *) string.c_str(), 
                          string.size());
#else
        // if there is no XftDrawStringUtf8, try to convert the
        // string to draw to locale encoding and use
        // XftDrawString8(). Probably this won't work for all languages
        // but at least it should work for languages which traditionally
        // used 8bit locales (e.g. European languages)
        XftDrawString8(draw, &c, _xftfont, x + _offset, 
                          _xftfont->ascent + y + _offset,
                          (XftChar8 *) utf8tolocale(string).c_str(), 
                          utf8tolocale(string).size());
#endif
      }
      else {
        // the string we want to draw isn't in UTF-8 but in
        // locale encoding.
#ifdef XFT_UTF8
        // convert to UTF-8 and draw with XftDrawStringUtf8()
        XftDrawStringUtf8(draw, &c, _xftfont, x + _offset, 
                       _xftfont->ascent + y + _offset,
                       (XftChar8 *) localetoutf8(string).c_str(),
                       localetoutf8(string).size());
#else
        // draw "as is" with XftDrawString8():
        XftDrawString8(draw, &c, _xftfont, x + _offset, 
                       _xftfont->ascent + y + _offset,
                       (XftChar8 *) string.c_str(), string.size());
#endif
      }
    }
    
    XftColor c;
    c.color.red = color.red() | color.red() << 8;
    c.color.green = color.green() | color.green() << 8;
    c.color.blue = color.blue() | color.blue() << 8;
    c.pixel = color.pixel();
    c.color.alpha = 0xff | 0xff << 8; // no transparency in BColor yet

    if (utf) {
      // the string we want to draw is in UTF-8.
#ifdef XFT_UTF8
      // draw it as is with XftDrawStringUtf8()
      XftDrawStringUtf8(draw, &c, _xftfont, x, _xftfont->ascent + y,
                        (XftChar8 *) string.c_str(), string.size());
#else
      // if there is no XftDrawStringUtf8, try to convert the
      // string to draw to locale encoding and use
      // XftDrawString8(). Probably this won't work for all languages
      // but at least it should work for languages which traditionally
      // used 8bit locales (e.g. European languages)
      XftDrawString8(draw, &c, _xftfont, x, _xftfont->ascent + y,
                     (XftChar8 *) utf8tolocale(string).c_str(),
                     utf8tolocale(string).size());
#endif
    }
    else {
      // the string we want to draw isn't in UTF-8 but in
      // locale encoding.
#ifdef XFT_UTF8
      // convert to UTF-8 and draw with XftDrawStringUtf8()
      XftDrawStringUtf8(draw, &c, _xftfont, x, _xftfont->ascent + y,
                        (XftChar8 *) localetoutf8(string).c_str(),
                        localetoutf8(string).size());
#else
      // draw "as is" with XftDrawString8():
      XftDrawString8(draw, &c, _xftfont, x, _xftfont->ascent + y,
                     (XftChar8 *) string.c_str(), string.size());
#endif
    }
    
    XftDrawDestroy(draw);
    return;
  }
#endif // XFT

  BPen pen(color);

  if (utf) {
    // the string we want to draw is in UTF-8.
#ifdef X_HAVE_UTF8_STRING
    // draw it as is with Xutf8DrawString()
    Xutf8DrawString(_display, d, _fontset, pen.gc(),
                    x, y - _fontset_extents->max_ink_extent.y,
                    string.c_str(), string.size());
#else
    // If Xutf8DrawString is not available, try to convert the string to
    // locale encoding and use XmbDrawString(). 
    XmbDrawString(_display, d, _fontset, pen.gc(),
                  x, y - _fontset_extents->max_ink_extent.y,
                  utf8tolocale(string).c_str(), utf8tolocale(string).size());
#endif
  }
  else {
    // the string we want to draw isn't in UTF-8 but in
    // locale encoding.
#ifdef X_HAVE_UTF8_STRING
    // convert to UTF-8 and draw with Xutf8DrawString()
    Xutf8DrawString(_display, d, _fontset, pen.gc(),
                    x, y - _fontset_extents->max_ink_extent.y,
                    localetoutf8(string).c_str(), localetoutf8(string).size());
#else
    // draw "as is" with XmbDrawString()
    XmbDrawString(_display, d, _fontset, pen.gc(),
                  x, y - _fontset_extents->max_ink_extent.y,
                  string.c_str(), string.size());
#endif
  }
   
}


unsigned int BFont::measureString(const string &string, bool utf) const {
  assert(_valid);

#ifdef    XFT
  if (_xftfont) {
    XGlyphInfo info;

    if (utf) {
      // the string to measure is UTF-8 encoded
#ifdef XFT_UTF8
      // measure it as is with XftTextExtentsUtf8()
      XftTextExtentsUtf8(_display, _xftfont, (XftChar8 *) string.c_str(),
                         string.size(), &info);
#else
      // if there is no XftTextExtentsUtf8(), try to convert to locale
      // encoding and use XftExtents8()
      XftTextExtents8(_display, _xftfont, (XftChar8 *) utf8tolocale(string).c_str(),
                      utf8tolocale(string).size(), &info);
#endif
    }
    else {
      // the string to measure isn't in UTF-8 but in locale encoding
#ifdef XFT_UTF8
      // convert to UTF-8 and measure with XftTextExtentsUtf8()
      XftTextExtentsUtf8(_display, _xftfont, (XftChar8 *) localetoutf8(string).c_str(),
                         localetoutf8(string).size(), &info);
#else
      // if there is no XftTextExtentsUtf8(), measure "as is" with XftTextExtents8()
      XftTextExtents8(_display, _xftfont, (XftChar8 *) string.c_str(),
                      string.size(), &info);
#endif
    }
    return info.xOff + (_shadow ? std::abs(_offset) : 0);
  }
#endif // XFT

  XRectangle ink, logical;

  if (utf) {
    // the string we want to measure is in UTF-8
#ifdef X_HAVE_UTF8_STRING
    // measure it as is with Xutf8TextExtents()
    Xutf8TextExtents(_fontset, string.c_str(), string.size(), &ink, &logical);
#else
    // If Xutf8TextExtents() is not available, try to convert the string to
    // locale encoding and use XmbTextExtents().
    XmbTextExtents(_fontset, utf8tolocale(string).c_str(), 
                   utf8tolocale(string).size(), &ink, &logical);
#endif
  }
  else {
    // the string we want to measure isn't in UTF-8 but in locale encoding
#ifdef X_HAVE_UTF8_STRING
    // convert to UTF-8 and measure with Xutf8TextExtents()
    Xutf8TextExtents(_fontset, localetoutf8(string).c_str(),
                     localetoutf8(string).size(), &ink, &logical);
#else
    // measure it "as is" with XmbTextExtents()
    XmbTextExtents(_fontset, string.c_str(), string.size(), &ink, &logical);
#endif
  }
  return logical.width;
}


unsigned int BFont::height(void) const {
  assert(_valid);

#ifdef    XFT
  if (_xftfont)
    return _xftfont->height + (_shadow ? std::abs(_offset) : 0);
#endif // XFT

    return _fontset_extents->max_ink_extent.height;
}


unsigned int BFont::maxCharWidth(void) const {
  assert(_valid);

#ifdef    XFT
  if (_xftfont)
    return _xftfont->max_advance_width;
#endif // XFT

    return _fontset_extents->max_logical_extent.width;
}
